stream_info = "WS:8090";


function readGecoItems( tango_dev )
{
    const web_url = tango_dev + "/items";
    let items = getTangoAttribute(web_url).value;
    return (items);
}

function populateGecoPanel( thePanel, tango_dev)
{
    // Read the items from GECO Tango Device
    let items = readGecoItems(tango_dev);
    var sorted_items = new Array(items.length);
    // Sort the items array as the DB MODBUS 
    for (var j = 0; j < items.length; j++) {
        // item format = "NAME:TYPE:ID"
        var item_tokens = items[j].split(":");
        sorted_items[item_tokens[2]-1] = item_tokens;
    }
    //
    const ValveTypes = ["VLV", "SHT", "STP", "SHP"];
    const AnalogTypes = ["TCS"];
    const AlarmTypes = ["FS","VGPE","SIP"];
    const AbiTypes = ["ABI"];
    const FastValveTypes = ["FVLV"];
    //
    var aInsp = donkiAddInspector(thePanel, {widgets_per_row: 3});
    //
    var last_type = "";
    for (var j = 0; j < sorted_items.length; j++) {
        var item_name = sorted_items[j][0];
        var item_type = sorted_items[j][1];
        // Add a separator between different set of widgets 
        if (last_type.length > 0) {
            if (last_type != item_type) donkiAddSeparator(aInsp);
        }
        last_type = item_type;
        //
        var item_tokens = item_name.split("_");
        // Add the correct widget to the panel 
        if (ValveTypes.includes(item_tokens[0])) {
            donkiAddLabel(aInsp, item_name, {width: 150, align: "right"});
            donkiAddReadSVG(aInsp, tango_dev + "/" + item_name, {size: 20,lut:lut_VALVE, shape:"rect"});
            donkiAddBoolControl(aInsp, null,tango_dev + "/" + item_name, ["Close","Open"],{width: 140});
            donkiAddReadSVG(aInsp, tango_dev + "/" + item_name+"_COIL", {size: 12,shape:"rect",lut:lut_ALARM});
            donkiAddReadSVG(aInsp, tango_dev + "/" + item_name+"_INTLCK", {size: 12,lut:lut_ALARM});
            donkiAddLineBreak(aInsp);
        } else if (FastValveTypes.includes(item_tokens[0])) {
            donkiAddLabel(aInsp, item_name, {width: 150, align: "right"});
            donkiAddReadSVG(aInsp, tango_dev + "/" + item_name, {size: 20,lut:lut_FASTVALVE, shape:"rect"});
            donkiAddLineBreak(aInsp);
        } else if (AnalogTypes.includes(item_tokens[0])) {
            donkiAddLabel(aInsp, item_name, {width: 150, align: "right"});
            donkiAddReadSVG(aInsp, tango_dev + "/" + item_name+"_ALARM", {size: 12,lut:lut_ALARM});
            donkiAddReadLabel(aInsp, tango_dev + "/" + item_name, {width: 100});
            donkiAddLineBreak(aInsp);
        } else if (AlarmTypes.includes(item_tokens[0])) {
            donkiAddLabel(aInsp, item_name, {width: 150, align: "right"});
            donkiAddReadSVG(aInsp, tango_dev + "/" + item_name+"_ALARM", {size: 12,lut:lut_ALARM});
            donkiAddLineBreak(aInsp);
        } else if (AbiTypes.includes(item_tokens[0])) {
            donkiAddLabel(aInsp, item_name, {width: 150, align: "right"});
            donkiAddReadSVG(aInsp, tango_dev + "/" + item_name, {size: 12,lut:lut_DEFAULT});
            donkiAddLineBreak(aInsp);
	} else {
            donkiAddLabel(aInsp, item_name, {width: 150, align: "right"});
            donkiAddLineBreak(aInsp);
        }
    }
}

function populateMyArea( mainarea )
{
    //
    if (typeof GECO_TD === "undefined") {
        console.log("GECO_TD not defined!");
        return;
    };

    // Create Main Panel
    var leftPanel = new LiteGUI.Panel("left_panel", {title:page_title, close:true,scroll:true});
    mainarea.add( leftPanel );
    // Create Right Panel
    var rightPanel = new LiteGUI.Panel("right_panel", {title:'Beamline Elements', close: true,scroll:true});
    mainarea.split("horizontal",[360,null],true);
    mainarea.getSection(1).add( rightPanel );

    //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Populate Left Panel
    var rInsp = donkiAddInspector(leftPanel, {widgets_per_row: 1});
    //
    donkiAddImage(rInsp,"/html/images/geco_spe_logo.png", {height:80});
    donkiAddLineBreak(rInsp);
    donkiAddLabel(rInsp, "ALARMS", {width: 150, align: "right"});
    donkiAddReadTextArea(rInsp, GECO_TD+"/"+"alarms", {height:80, width: 350});
    donkiAddLabel(rInsp, "INTERLOCKS", {width: 150, align: "right"});
    donkiAddReadTextArea(rInsp, GECO_TD+"/"+"interlocks", {height:80, width: 350});
    donkiAddLabel(rInsp, "BYPASSED", {width: 150, align: "right"});
    donkiAddReadTextArea(rInsp, GECO_TD+"/"+"bypassed", {height:80, width: 350});
    donkiAddLabel(rInsp, "LOG", {width: 150, align: "right"});
    donkiAddReadTextArea(rInsp, GECO_TD+"/"+"logs", {height:300, width: 350, reverse: true});

    //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Populate Right Panel
    populateGecoPanel(rightPanel, GECO_TD);
}
