var elem_ids = {};
var past_values = {};
var page_title = "MAIN PANEL";

var mainarea = null;

var chunkSize = 30;
var readPeriod_ms = 500;
var write_attrs_history = { "wattrs":[], "wtimes":[] };

const INT_TYPES = [1,2,3,6,7,22,23,24,27];
const FLOAT_TYPES = [4,5];
const STRING_TYPES = [8];
const SPECTRUM_TYPES = [9,10,11,12,13,14,15,16,25,26];
const STATE_TYPES = [19]
const SUPPORTED_TYPES = INT_TYPES.concat(FLOAT_TYPES, STRING_TYPES, SPECTRUM_TYPES, STATE_TYPES);

const lut_FASTVALVE = {'colors': {0:'lime',1:'orangered',2:'navy',3:'orange' }};
const lut_STATE = {'colors': {0:'lime',1:'white',2:'white',3:'white',4:'white',5:'lime',6:'blue',7:'yellow',8:'orangered',9:'beige',10:'lime',11:'orange',12:'magenta',13:'grey'}};
const lut_VALVE = {'colors': {0:'lime',1:'orangered',2: 'navy', 3:'orange' }, 'choices': {'Close':0,'Open':1}};
const lut_MACHINEVALVE = {'colors': {1:'orangered',2:'lime',9:'orangered',10:'lime',0:'orange',3:'orange',4:'orange',5:'orange',6:'orange',7:'orange',8:'orange'}};
const lut_ALARM = {'colors': {0:'lime',1:'orangered'}};
const lut_ONLYALARM = {'colors': {0:'none',1:'orange'}};
const lut_INTLCK = {'colors': {0:'lime',1:'orangered'}};
const lut_LIGHT = {'colors': {0:'black',1:'yellow'}};
const lut_BOOL = {'colors': {0:'orangered',1:'lime'}, 'choices': {'False':0,'True':1}};
const lut_DEFAULT = {'colors': {0:'darkgray',1:'lime'}, 'choices': {'Off':0,'On':1}};

/*
 Create an utility object that connects each control point to the page elements  
*/
var source_points = {};
var stream_info = "";  // Could be override in the user script
var dialog_input = null;

function update_source_points()
{
    for (var member in source_points) delete source_points[member];
    let el_list = Object.keys(elem_ids);
    for (let i = 0; i < el_list.length; i += 1) {
        let e_id = el_list[i];
        if (e_id.startsWith("dButton")) {
            continue;
        };
        if (Array.isArray(elem_ids[e_id])) {
            c_points = elem_ids[e_id];
        } else {
            c_points = [elem_ids[e_id]];
        }
        for (let i = 0; i < c_points.length; i += 1) {
            let c_point = c_points[i];
            if (c_point in source_points) {
                source_points[c_point].push(e_id);
            } else {
                source_points[c_point] = [e_id];
            }
        };
    };
}

window.onload = function(){

    LiteGUI.init(); 

    mainarea = new LiteGUI.Area({ id: "mainarea", content_id:"canvasarea", height: "calc( 100% - 20px )", main:true, inmediateResize: true});
    LiteGUI.add( mainarea );
    populateMyArea( mainarea );
    update_source_points();
    if ((stream_info == "SSE") && (typeof(EventSource) !== "undefined"))  {
        // USE Server Side Events
        useEventSources();
    } else if (stream_info.startsWith("WS")){
        // USE WebSocket
        const WS_params = stream_info.split(":");
        if (WS_params.length > 1) {
            ws_url = "ws://"+location.hostname+":"+WS_params[1];
        } else {
            ws_url = "ws://"+location.hostname+":8090";
        }
        useWebSocketSources(ws_url);
    } else {
        // USE periodic polling
        var intervalId = setInterval(periodicReadSources, readPeriod_ms);
    }
};


function periodicReadSources()
{
    let req = new XMLHttpRequest();
    let source_list = Object.keys(source_points);
    for (let i = 0; i < source_list.length; i += chunkSize) {
        const source_chunk = source_list.slice(i, i + chunkSize);
        const web_url = "/tango-cs/GETATTR?td=" + source_chunk;
        req.open('GET', web_url, false);
        req.send(null);
        if(req.status == 200) {
            parseResponseData(req.responseText);
        };
    };
};

function useWebSocketSources(stream_url)
{
    var source = new WebSocket(stream_url);
    source.onmessage = function(event){
        parseResponseData(event.data);
    }
    source.onopen = function(){
        source.send("period="+readPeriod_ms);
        let source_list = Object.keys(source_points);
        for (let i = 0; i < source_list.length; i += chunkSize) {
            const source_chunk = source_list.slice(i, i + chunkSize);
            source.send("td=" + source_chunk);
        };
        setTimeout(doStartStreaming, 500);
    }
    function doStartStreaming() {
        source.send("cmd=start_streaming");
    }
    source.onclose = function(){
        location.reload()
    }
    source.onerror = function(){
        LiteGUI.alert("Websocket error", {title: "Websocket error"});
    }

};

function useEventSources()
{
    let source_list = Object.keys(source_points);
    for (let i = 0; i < source_list.length; i += chunkSize) {
        const source_chunk = source_list.slice(i, i + chunkSize);
        var source = new EventSource("/tango-cs/STREAM?period="+readPeriod_ms+"&td=" + source_chunk);
        source.onmessage = function(event) {
            parseResponseData(event.data);
        };
    };
};

function onWriteSuccess(reply, request)
{
    if (typeof reply === 'string') { 
        response = JSON.parse(reply);
    } else {
        response = reply;
    };
    let caption = response.reply + "<BR/>";
    if (Object.keys(response).includes("retval")) {
        caption = caption + "Output: " + response.retval;
    }
    LiteGUI.alert( caption, {title: response.name});
};

function onWriteError(errormsg)
{
    LiteGUI.alert(errormsg, {title: "Write Error"});
};

function getTangoInfo(tgpoint, type = "cmd")
{
    let req = new XMLHttpRequest();
    let web_url;
    if (type == "cmd") {
        web_url = "/tango-cs/CMDINFO?td=" + tgpoint;
    } else {
        web_url = "/tango-cs/ATTRINFO?td=" + tgpoint;
    };
    req.open('GET', web_url, false);
    req.send(null);
    if(req.status != 200) {
        onWriteError(req.status +" "+req.statusText);
        return;
    };
    let dInfo = JSON.parse(req.responseText);
    return dInfo;
};

function getTangoAttribute(tgname, force = false)
{
    let req = new XMLHttpRequest();
    let web_url = "/tango-cs/GETATTR?td=" + tgname;
    if (force) web_url += "&force=true"
    req.open('GET', web_url, false);
    req.send(null);
    if(req.status != 200) {
        onWriteError(req.status +" "+req.statusText);
        return null;
    };
    let attr_result = JSON.parse(req.responseText);
    return attr_result[0];
};

function readTangoAttribute(tgname, need_reply = true)
{
    const web_url = "/tango-cs/GETATTR?td=" + tgname;
    console.log(web_url);
    if (need_reply) {
        LiteGUI.requestText(web_url, onWriteSuccess, onWriteError);
    } else {
        LiteGUI.requestText(web_url, null, onWriteError);
    }
};

function writeTangoAttribute(tgname, value, need_reply = true)
{
    const web_url = "/tango-cs/SETATTR?td=" + tgname +"," + value;
    let tnow = Date.now();
    let aidx = write_attrs_history.wattrs.indexOf(tgname);
    if (aidx > -1) {
        let tdiff = Date.now() - write_attrs_history.wtimes[aidx];
        // limit the frequency of the write operation on the same attribute
        if (tdiff < 2000) return;
        write_attrs_history.wtimes[aidx] = tnow;
    } else {
        write_attrs_history.wattrs.unshift(tgname);
        write_attrs_history.wtimes.unshift(tnow);
        write_attrs_history.wattrs.length = Math.min(write_attrs_history.wattrs.length, 3);
        write_attrs_history.wtimes.length = Math.min(write_attrs_history.wtimes.length, 3);
    };
    //
    console.log("writeTangoAttribute "+ web_url);
    if (need_reply) {
        LiteGUI.requestText(web_url, onWriteSuccess, onWriteError);
    } else {
        LiteGUI.requestText(web_url, null, onWriteError);
    }
};

// Force a Tango attribute readout
function force_update_value(attrname) {
    // This method could be called by a setTimeout
    var attrval = getTangoAttribute(attrname, true);
    parseResponseData(JSON.stringify([attrval]));
}

// Utility method to Enable/Disable a widget
function set_widget_enabled(widget, value_in) {
    // This method could be called by a setTimeout
    widget.set_enabled(value_in);
}


function createInputParamDialog(tginfo, type = "cmd", need_reply = true, parent = null) {
    // Create a new dialog
    tgname = tginfo.name
    var inputValue;
    // Close previous opened dialogs
    if (dialog_input != null) dialog_input.close();
    //
    var options = { title:tgname, close: true, minimize: false};
    options.width = 300;
    options.height = 60;
    options.scroll = false;
    options.resizable = false;
    options.draggable = true;
    options.parent = parent;
    var dialog = new LiteGUI.Dialog(tgname, options);
    dialog_input = dialog;
    dialog.adjustSize();
    if (parent != null) {
        let rect = parent.getBoundingClientRect();
        dialog.setPosition(rect.left, rect.top);
    };
    // Create a collection of widgets
    var widgets = new LiteGUI.Inspector();
    //var nameWidget = widgets.addString("Your name","foo");
    if ((type == "cmd")) {
        dtype = tginfo.in_type;
    } else {
        dtype = tginfo.data_type;
    };
    prec_in = 2
    if (INT_TYPES.includes(dtype)) {
        prec_in = 0;
    }
    var numberWidget = widgets.addNumber("Input", value = tginfo.value, {precision : prec_in});
    dialog.add(widgets);
    //
    function applyMethod() {
        inputValue = numberWidget.getValue();
        if (type == "cmd") {
            sendCommand(tgname, inputValue, need_reply = need_reply);
        } else {
            writeTangoAttribute(tgname, inputValue, need_reply = need_reply);
        };
    }
    //
    function keyPressed( e ) {
        if (e.key === "Enter") {
            dialog.close();
            applyMethod();
        };
    }
    // Add some buttons
    dialog.addButton('Apply', { close: 'fade', callback: applyMethod });
    dialog.addButton('Cancel',{ close: 'fade' });
    numberWidget.addEventListener("keypress", keyPressed);
    return dialog;
}


function createInputBooleanDialog(tgname,true_text="Close",false_text="Open", parent = null) {
    let req = new XMLHttpRequest();
    let web_url = "/tango-cs/GETATTR?td=" + tgname;
    req.open('GET', web_url, false);
    req.send(null);
    if(req.status != 200) {
        onWriteError(req.status +" "+req.statusText);
        return;
    };
    let attr_result = JSON.parse(req.responseText);
    attr_value = attr_result[0].value;
    var inputValue;
    // Close previous opened dialogs
    if (dialog_input != null) dialog_input.close();
    // Create a new dialog
    var options = { title:tgname, close: true, minimize: false};
    options.width = 300;
    options.height = 60;
    options.scroll = false;
    options.resizable = false;
    options.draggable = true;
    options.parent = parent;
    var dialog = new LiteGUI.Dialog(tgname, options);
    dialog_input = dialog;
    if (parent != null) {
        let rect = parent.getBoundingClientRect();
        dialog.setPosition(rect.top, rect.left);
    };
    //
    function applyMethod() {
        writeTangoAttribute(tgname, inputValue, need_reply = true);
    }
    // Add some buttons
    if (attr_value) {
        aButt = dialog.addButton(true_text, { close: 'fade', callback: applyMethod });
        inputValue = 0;
    } else {
        aButt = dialog.addButton(false_text, { close: 'fade', callback: applyMethod });
        inputValue = 1;
    }
    dialog.addButton('Cancel',{ close: 'fade' });
    return dialog;
}


function synoptic_click(ele) {
    lut = ele.getAttribute("lut");
    if (lut == undefined) lut = lut_DEFAULT;
    let tgname = elem_ids[ele.id];
    options = {};
    options.title = "SET OPERATION";
    options.width = 400;
    obj_choices = JSON.parse(lut).choices;
    choices = Object.keys(obj_choices);
    choices.push("Cancel");
    let arginDialog = LiteGUI.choice( tgname , choices, inner, options);
    //
    function inner(v) {
        if (v == "Cancel") return;
        inputValue = obj_choices[v];
        writeTangoAttribute(tgname, inputValue, need_reply = true);
    };
};

function sendCommand(tgname, value, need_reply = true)
{
    let req = new XMLHttpRequest();
    let web_url = "/tango-cs/CMD?td=" + tgname;
    if (value !== null) {
        web_url = web_url +"," + value;
    };
    console.log("sendCommand "+ web_url);
    if (need_reply) {
        LiteGUI.requestJSON(web_url, onWriteSuccess, onWriteError);
    } else {
        LiteGUI.requestJSON(web_url, null, onWriteError);
    };
};

function askInfo(id, value, type = "cmd", need_reply = true, parent = null)
{
    let unknown_el = !(id in elem_ids);
    if (unknown_el) {
        return;
    };
    let req = new XMLHttpRequest();
    let web_url;
    if (type == "cmd") {
        web_url = "/tango-cs/CMDINFO?td=" + elem_ids[id];
    } else {
        web_url = "/tango-cs/ATTRINFO?td=" + elem_ids[id];
    };
    req.open('GET', web_url, false);
    req.send(null);
    if(req.status != 200) {
        onWriteError(req.status +" "+req.statusText);
        return;
    };
    let dInfo = JSON.parse(req.responseText);
    dInfo.value = value;
    if (type == "cmd") {
        // Manage commands
        if (SUPPORTED_TYPES.includes(dInfo.in_type)) {
            if (value !== null) {
                sendCommand(elem_ids[id], value = value, need_reply);
            } else {
                let arginDialog = createInputParamDialog(dInfo, type, parent);
                arginDialog.show('fade');
            };
        } else if (dInfo.in_type == 0) {
            sendCommand(elem_ids[id], value = null, need_reply);
        } else  {
            onWriteError("Data type not supported");
        };
    } else {
        // Manage attributes
        if (SUPPORTED_TYPES.includes(dInfo.data_type)) {
            let arginDialog = createInputParamDialog(dInfo, type, need_reply, parent);
            arginDialog.show('fade');
        } else  {
            onWriteError("Data type not supported");
        };
    };
};

function _rwlabel_inputClicked() {
    askInfo(this.id, this.getValue(), type = "attr", need_reply = this.options.need_reply, parent = this);
};

function parseResponseData( json_data )
{
    const d = new Date();
    let time = d.getTime();
    const objs = JSON.parse(json_data);
    for (let idx = 0; idx < objs.length; idx += 1) {
        obj = objs[idx];
        //
        if ((past_values.hasOwnProperty(obj.name)) && (past_values[obj.name] == obj.value)) {
            continue;
        }
        //
        past_values[obj.name] = obj.value;
        txt_val = "";
        fillcolor = "red";
        if (obj.value == null) {
            txt_val = "---";
            fillcolor = "black";
        } else if (Array.isArray(obj.value))  {
            txt_val = obj.value.join("\n");
        } else if ((obj.type == 4) || (obj.type == 5))  {
            if ((Math.abs(obj.value) < 0.01) || (Math.abs(obj.value) > 9999)) {
                txt_val = obj.value.toExponential(2);
            } else {
                txt_val = obj.value.toFixed(2);
            }
        } else if ((obj.type == 1) || (obj.type == 2)) {
            txt_val = obj.value;
            if (Number(obj.value) == 1) {
                fillcolor = "red";
            } else if (Number(obj.value) == 0) {
                fillcolor = "gray";
            } else {
                fillcolor = "navy";
            };
        } else {
            txt_val = obj.value;
        };
        var elements_id = source_points[obj.name];
        if (elements_id == undefined) continue;
        //
        for (let i = 0; i < elements_id.length; i++) {
            if (document.getElementById(elements_id[i]) == undefined) continue;
            // Manage postfix
            if (document.getElementById(elements_id[i]).hasAttribute("postfix")) {
                postfix = " " + document.getElementById(elements_id[i]).getAttribute("postfix");
            } else if (obj.unit != "") {
                postfix = " " +  obj.unit;
            } else { postfix = "";};
            txt_to_show = (txt_val + postfix).trim()
            //
            if (elements_id[i].startsWith("dLabel") || elements_id[i].startsWith("dWlabel")) {
                // Update Labels
                document.getElementById(elements_id[i]).setValue((txt_to_show), skip_event = true);
            } else if (elements_id[i].startsWith("dTextArea")) {
                let txtA = document.getElementById(elements_id[i]);
                let innerTextArea = txtA.querySelector(".wcontent textarea");
                let prevScrollHeight = innerTextArea.scrollHeight;
                txtA.setValue((txt_val), skip_event = true);
                if ((txtA.getAttribute("reverse")) && (prevScrollHeight < innerTextArea.scrollHeight)) {
                    innerTextArea.scrollTop = innerTextArea.scrollHeight;
                }
            } else if (elements_id[i].startsWith("dSvgColor")) {
                // Update SVG color
                fillcolor = "black";
                if (obj.value != null) {
                    let lut = JSON.parse(document.getElementById(elements_id[i]).getAttribute('lut'));
                    fillcolor = lut.colors[Number(obj.value)];
                }
                document.getElementById(elements_id[i]).style.fill = fillcolor;
            } else if (elements_id[i].startsWith("dColorLabel")) {
                // Update font color
                if (obj.value != null) {
                    let lut = JSON.parse(document.getElementById(elements_id[i]).getAttribute('lut'));
                    fontcolor = lut.colors[Number(obj.value)];
                    document.getElementById(elements_id[i]).style.color = fontcolor;
                }
            } else if (elements_id[i].startsWith("dSvgLabel")) {
                // Update SVG text
                document.getElementById(elements_id[i]).textContent = txt_to_show;
            } else if (elements_id[i].startsWith("dCheckbox")) {
                // Update CheckBox Value
                if (obj.value !== document.getElementById(elements_id[i]).getValue()) {
                    let user_click = document.getElementById(elements_id[i]).getAttribute("actioned");
                    document.getElementById(elements_id[i]).setAttribute("actioned", false);
                    if (user_click !== "true") {
                        // This is for skipping the first update cycle after a user click
                        document.getElementById(elements_id[i]).click();
                    };
                };
            } else if (elements_id[i].startsWith("dBoolButtons")) {
                // Update Boolean Control Buttons
            } else if (elements_id[i].startsWith("dDisableCondition")) {
                // Update Enable/Disable conditions
                let eWdg = document.getElementById(elements_id[i]);
                eWdg.apply_condition(txt_val);
            } else if (elements_id[i].startsWith("dPlotLine")){
                document.getElementById(elements_id[i]).newValue(obj);
            } else if (elements_id[i].startsWith("dSynoptic")){
                // Update Synoptic element
                let aSVG = document.getElementById(elements_id[i])
                if (typeof aSVG.getSVGDocument !== 'function') {
                    continue;
                    }
                let doc = aSVG.getSVGDocument();
                if (doc == null) {continue;}
                let lut = JSON.parse(aSVG.getAttribute("lut"));
                let syn_elems = doc.getElementsByClassName("synoptic_element");
                for (let synEl of syn_elems) {
                    fillcolor = lut.colors[Number(obj.value)];
                    synEl.style.fill = fillcolor;
                };
            } else {
                // By default update inner Text
                document.getElementById(elements_id[i]).innerText = txt_to_show;
            };
        };
    };
};

function aWidgetChanged(name,value,widget) {
    let unknown_el = !(widget.id in elem_ids);
    if (unknown_el) {
        return;
    };
    if (widget.id.startsWith("dCheckbox") || (widget.id.startsWith("dBoolButtons"))) {
        user_action = document.getElementById(widget.id).getAttribute("actioned");
        if ( user_action == "true") {
            document.getElementById(widget.id).setAttribute("actioned", true);
            if ( widget.id.startsWith("dBoolButtons") ) {
                var buttons = document.getElementById(widget.id).querySelectorAll("button");
                for(var i = 0; i < buttons.length; ++i) {
                    if ((value) == buttons[i].innerHTML) {
                        value = i;
                        widget.applyclick(value);
                        return;
                    };
                };
            };
            writeTangoAttribute(elem_ids[widget.id], (value ? 1 : 0), need_reply = true);
        };
        document.getElementById(widget.id).setAttribute("actioned", true);
    } else if (widget.id.startsWith("dButton")) {
        return;
    } else {
        console.log("A Widget changed: " + name + " -> " + value);
    };
}

function _util_set_element_grid_style( element, options)
{
    let row_info = "";
    let col_info = "";
    if (options.row) row_info += options.row;
    if (options.row_span) row_info += "/span "+ options.row_span;
    if (options.col) col_info += options.col;
    if (options.col_span) col_info += "/span "+ options.col_span;
    element.style.gridRow = row_info;
    element.style.gridColumn = col_info;
    return element;
}


function donkiAddInspector( parent, options = {})
{
    var widgets = new LiteGUI.Inspector(options);
    widgets.onchange = aWidgetChanged;
    parent.add( widgets );
    return widgets
}

function donkiAddGridInspector( parent, options = {})
{
    var widgets = new LiteGUI.Inspector(options);
    widgets.onchange = aWidgetChanged;
    widgets.root.style.display = "inline-grid";
    widgets.root.style.gridTemplateColumns = "auto";
    widgets.root.style.alignItems = "center";
    widgets.root.style.justifyContent = "center";
    parent.add( widgets );
    return widgets
}

function donkiAddLineBreak( parent ) 
{
    aBrk = parent.addSeparator();
    aBrk.className = null;
    return aBrk;
}

function donkiAddSeparator( parent ) 
{
    return parent.addSeparator();
}

function donkiAddSection( parent , name, options = {}) 
{
    return parent.addSection(name, options);
}

function donkiAddTabsContainer( parent , options = {}) 
{
    /**
    * Widget that contains several tabs and their content
    * Options:
    * - mode: "vertical","horizontal"
    * - size
    * - width,height
    * - autoswitch: allow autoswitch (switch when mouse over)
    * @class Tabs
    * @constructor
    */
    var tabs_widget = new LiteGUI.Tabs(options);
    parent.add( tabs_widget );
    return tabs_widget;

}
        
function donkiAddImage( parent, image_url, options = {})
{
    options.disabled = true;
    var aIcon = parent.addInfo(null,null, options);
    var options_img = "";
    if(options.width)
        options_img += "width=" + options.width;
    if(options.height)
        options_img += "height=" + options.height;
    aIcon.innerHTML = "<span class='winfo'><img src='" + image_url + "' " + options_img + "/></span>";
    aIcon.style.display = "inline-block";
    aIcon = _util_set_element_grid_style(aIcon, options);
    return aIcon
}

function donkiAddLabel( parent, caption, options = {})
{
    options.disabled = true;
    aLabel = parent.addInfo(null, caption, options);
    if (options.align) aLabel.style.textAlign = options.align;
    if (options.color) aLabel.style.color = options.color;
    aLabel = _util_set_element_grid_style(aLabel, options);
    return aLabel
}

function donkiAddReadLabel( parent, attrname, options = {}) 
{
    let el_list = Object.keys(elem_ids);
    let el_name = "dLabel" + el_list.length;
    options.disabled = true;
    caption = null;
    if (options.caption != undefined) caption = options.caption;
    aS = parent.addString(caption, "-", options);
    aS.setAttribute("id", el_name);
    aS.setAttribute("title", attrname);
    elem_ids[el_name] = attrname;
    if (options.postfix != undefined) {
        aS.setAttribute("postfix", options.postfix);
    }
    aS = _util_set_element_grid_style(aS, options);
    return aS;
}

function donkiAddReadColorLabel( parent, attrname, caption, options = {}) 
{
    let el_list = Object.keys(elem_ids);
    let el_name = "dColorLabel" + el_list.length;
    options.disabled = true;
    var lut = lut_DEFAULT;
    if (options.lut) lut = options.lut;
    let text_color = "white";
    let lut_attr = JSON.stringify(lut);
    colorLabel = parent.addInfo(null, caption, options);
    colorLabel.setAttribute("id", el_name);
    colorLabel.setAttribute("lut", lut_attr);
    if (options.align) colorLabel.style.textAlign = options.align;
    if (options.color) colorLabel.style.color = options.color;
    colorLabel = _util_set_element_grid_style(colorLabel, options);
    elem_ids[el_name] = attrname;
    return colorLabel
}

function donkiAddReadWriteLabel( parent, caption, attrname, options = {}) 
{
    let el_list = Object.keys(elem_ids);
    let el_name = "dWlabel" + el_list.length;
    options.disabled = false;
    aWlbl = parent.addString(caption, "-", options);
    aWlbl.setAttribute("title", attrname);
    aWlbl.setAttribute("id", el_name);
    aWlbl_input = aWlbl.querySelector(".wcontent input");
    aWlbl_input.style.cursor = "pointer";
    aWlbl.addEventListener("dblclick", _rwlabel_inputClicked);
    //aWlbl.addEventListener("click", _rwlabel_inputClicked);
    elem_ids[el_name] = attrname;
    if (options.postfix != undefined) {
        aWlbl.setAttribute("postfix", options.postfix);
    }
    if (options.need_reply == undefined) options.need_reply = false;
    aWlbl.setAttribute("need_reply", options.need_reply);
    //
    aWlbl = _util_set_element_grid_style(aWlbl, options);
    //
    aWlbl.set_enabled = function(value_in) {
        if (value_in) {
            this.enable();
        } else {
            this.disable();
        }
    };
    return aWlbl;
}

function donkiAddReadTextArea( parent , attrname, options = {}) 
{
    let el_list = Object.keys(elem_ids);
    let el_name = "dTextArea" + el_list.length;
    options.disabled = true;
    aTA = parent.addTextarea(null, "", options);
    aTA.setAttribute("id", el_name);
    aTA.setAttribute("title", attrname);
    elem_ids[el_name] = attrname;
    if (options.reverse) {
        aTA.setAttribute("reverse", options.reverse);
        let innerTextArea = aTA.querySelector(".wcontent textarea");
        innerTextArea.scrollTop = innerTextArea.scrollHeight;
    }
    aTA = _util_set_element_grid_style(aTA, options);
    return aTA;
}

function donkiAddCheckBox( parent , caption, attrname, options = {}) 
{
    let el_list = Object.keys(elem_ids);
    let el_name = "dCheckbox" + el_list.length;
    options.callback = null;
    aChkBx = parent.addCheckbox(caption,true,options);
    aChkBx.setAttribute("id", el_name);
    aChkBx.setAttribute("actioned", true);
    aChkBx.setAttribute("title", attrname);
    elem_ids[el_name] = attrname;
    aChkBx = _util_set_element_grid_style(aChkBx, options);
    //
    aChkBx.set_enabled = function(value_in) {
        console.log("set_enabled todo");
    };
    return aChkBx;
}


function donkiAddBoolControl( parent , caption, attrname, labels, options = {}) 
{
    /**
    * Widget that create a Button for sending a Tango Command
    * Options:
    * - width,height
    * - confirm: enable confirmation popup (boolean)
    */
    let el_list = Object.keys(elem_ids);
    let el_name = "dBoolButtons" + el_list.length;
    options.callback = this.wclick;
    btnLabels = ["False","True"];
    if (labels) {
        btnLabels = labels;
        };
    aBtns = parent.addButtons(caption, btnLabels, options);
    aBtns.setAttribute("actioned", true);
    aBtns.setAttribute("id", el_name);
    aBtns.setAttribute("title", attrname);
    aBtns.options = options;
    aBtns.labels = labels;
    elem_ids[el_name] = attrname;
    aBtns = _util_set_element_grid_style(aBtns, options);
    //
    aBtns.set_enabled = function(value_in) {
        let buttons = this.querySelectorAll("button");
        for(var i = 0; i < buttons.length; ++i) {
            buttons[i].disabled = !value_in;
        };
    };
    //
    aBtns.applyclick = function (value){
        if (options.confirm === true) {
            var dlg_options = { title:"CONFIRM", close: true, minimize: false};
            let choices = ["Yes","No"];
            let caption = "<center>You pressed "+this.labels[value]+".<BR>Do you want to proceed?<BR></center>" ;
            let dialog = LiteGUI.choice( caption, choices, writeAttr, dlg_options);
        } else {
            writeTangoAttribute(elem_ids[el_name], (value ? 1 : 0), need_reply = true);
            return;        
        }
        //
        function writeAttr(v) {
            if (v == "No") return;
            writeTangoAttribute(elem_ids[el_name], (value ? 1 : 0), need_reply = true);
        };
    };
    return aBtns;
}

function donkiAddCommandButton( parent, caption, cmdname, options = {}) 
{
    /**
    * Widget that create a Button for sending a Tango Command
    * Options:
    * - width,height
    * - confirm: enable confirmation popup (boolean)
    */
    let el_list = Object.keys(elem_ids);
    let el_name = "dButton" + el_list.length;
    options.callback = null;
    aBtn = parent.addButton(null, caption, options);
    aBtn.setAttribute("id", el_name);
    aBtn.caption = caption;
    aBtn.options = options;
    elem_ids[el_name] = cmdname;
    aBtn = _util_set_element_grid_style(aBtn, options);
    //
    aBtn.set_enabled = function(value_in) {
        if (value_in) {
            this.enable();
        } else {
            this.disable();
        }
    };
    //
    aBtn.onclick = function (){
        if (options.confirm === true) {
            var dlg_options = { title:"CONFIRM", close: true, minimize: false};
            let choices = ["Yes","No"];
            let caption = "<center>Are you sure you wanna send<BR>"+aBtn.caption+"<BR>command?<BR></center>" ;
            let dialog = LiteGUI.choice( caption, choices, applyCommand, dlg_options);
        } else {
            askInfo(el_name, null, type = "cmd");
            return;        
        }
        //
        function applyCommand(v) {
            if (v == "No") return;
            askInfo(el_name, null, type = "cmd");
        };
    };
    return aBtn;
}

function donkiAddLinePlot( parent, caption, attr_x, attr_y, options = {}) 
{
    let el_list = Object.keys(elem_ids);
    let el_name = "dPlotLine" + el_list.length;
    var chart_size = options.chart_size || 500;
    var element = document.createElement("div");
    element.style.minHeight = "50px";
    element.style.width = options.width || "100%";
    element.options = options;
    element.style.minWidth = "50px";
    element.style.minHeight = "20px";
    element = _util_set_element_grid_style(element, options);
    var plot_canvas = document.createElement("canvas");
    plot_canvas.width = options.width || 200;
    plot_canvas.height = options.height || 50;
    element.appendChild( plot_canvas );
    element.canvas = plot_canvas;
    var x_sources = null;
    var y_sources = null;
    var linecolor = options.linecolor || "orange";
    var tmp_newdata = {X:[],Y:[]};
    //
    this.xAxisFormatter = function(val) {
        if (data.labels.length < 10) return val;
        let delta = Math.round(data.labels.length / 10);
        let index = data.labels.indexOf(val);
        if ((index % delta) !== 0 ) return "";
        return val;
    }
    //
    var config = {
        canvas: plot_canvas,
        axisLineColor: "gray",
        axisTextColor: "gray",
        yAxisMin: null,
        yAxisMax: null,
        datasetColorPicker: function (dataset) { return linecolor;},
        datasetPointSizePicker: function (dataset) { return 1;},
        xAxisFormatter: this.xAxisFormatter
    };
    //
    var aPlot = new SlimChart(config);
    var data = {
         labels: [],
         datasets: [{data: []}]
    };
    //console.log(aPlot.config);
    if (attr_y !== null) {
        element.setAttribute("id", el_name);
        elem_ids[el_name] = [attr_y];
        y_sources = attr_y;
        if (attr_x !== null) {
            elem_ids[el_name].push(attr_x);
            x_sources = attr_x;
        }
    }
    //
    function getMinMax(arr) {
        let min = arr[0];
        let max = arr[0];
        let i = arr.length;
        while (i--) {
            min = arr[i] < min ? arr[i] : min;
            max = arr[i] > max ? arr[i] : max;
        }
        return { min, max };
    }
    // newValue() method called for any new value
    element.newValue = function(new_obj) {
        if (new_obj.has_failed) return;
        let ddim = new_obj.r_dimension;
        if ((ddim == undefined) || (ddim[1] > 0)) return;
        if (ddim[0] == 1) {
            // Scalar
            if (new_obj.name === y_sources) {
                tmp_newdata.Y = [Number(new_obj.value)];
                if (x_sources === null) {
                    tmp_newdata.X = [Number(data.labels.slice(-1)) + 1];
                }
            } else if (new_obj.name === x_sources) {
                if ((new_obj.type == 4) || (new_obj.type == 5))  {
                    if ((Math.abs(new_obj.value) < 0.01) || (Math.abs(new_obj.value) > 9999)) {
                        txt_val = new_obj.value.toExponential(2);
                    } else {
                        txt_val = new_obj.value.toFixed(2);
                    }
                } else txt_val = Number(new_obj.value);
                tmp_newdata.X = [txt_val];
            }
            if ((tmp_newdata.X.length > 0) && (tmp_newdata.Y.length > 0)) {
                if (data.labels.length > 1) {
                    try {
                        let x0 = Number(data.labels.slice(-2)[0]);
                        let x1 = Number(data.labels.slice(-2)[1]);
                        let xnow = Number(tmp_newdata.X[0]);
                        if (((x1 >= x0) && (x1 > xnow)) || ((x1 <= x0) && (x1 < xnow))) {
                            // Reset the plot, inverted X direction
                            data.labels = [];
                            data.datasets[0].data = [];
                        }
                    } catch (e) {
                        console.log(e);
                    }
                }
                data.datasets[0].data.push(tmp_newdata.Y[0]);
                data.labels.push(tmp_newdata.X[0]);
                tmp_newdata.X = [];
                tmp_newdata.Y = [];
            } else 
                return;
            if (data.labels.length > chart_size) {
                data.labels.shift();
                data.datasets[0].data.shift();
            }
        } else {
            // Spectrum
            if (new_obj.name === y_sources) {
                tmp_newdata.Y = new_obj.value;
                if (x_sources === null) {
                    tmp_newdata.X = Array.from({ length: ddim[0] }, (v, i) => i);
                }
            } else if (new_obj.name === x_sources) {
                tmp_newdata.X = new_obj.value;
            }
            if (tmp_newdata.X.length !== tmp_newdata.Y.length )
                return
            data.labels = tmp_newdata.X;
            data.datasets[0].data = tmp_newdata.Y;
        }
        yylimits = getMinMax(data.datasets[0].data);
        if ((aPlot.config.yAxisMin === null) || (yylimits.min < aPlot.config.yAxisMin))
            aPlot.config.yAxisMin = yylimits.min;
        if ((aPlot.config.yAxisMax === null) || (yylimits.max > aPlot.config.yAxisMax))
            aPlot.config.yAxisMax = yylimits.max;
        aPlot.draw(data);
    }
    parent.append(element);
    return element;
}

function donkiAddReadSVG( parent , attrname, options = {}) 
{
    let el_list = Object.keys(elem_ids);
    let el_name = "dSvgColor" + el_list.length;
    let size = 30;
    let SHAPES = ["circle", "rect"];
    var lut = lut_DEFAULT;
    var shape = "circle";
    var caption = "";
    if (options.caption) caption = options.caption;
    if (options.lut) lut = options.lut;
    if (options.size) size = options.size;
    if ((options.shape) && (SHAPES.includes(options.shape))) {
        shape = options.shape;
    }
    let stroke_opt = "stroke=black stroke-width=2";
    let lut_attr = "lut=" + JSON.stringify(lut);
    halfSize = Math.floor(size / 2);
    ray = halfSize * 0.9;
    if (caption.length > 0) parent.addInfo(null,caption, {width: (caption.length * 9)});
    aSVG = parent.addInfo(null,"", {width: (size+10), height: size});
    aSVG = _util_set_element_grid_style(aSVG, options);
    svgCode = "<svg height="+size+" width="+size+"><title>"+attrname+"</title>";
    if (shape == "circle") {
         svgCode += "<circle id='"+el_name+"' cx="+halfSize+" cy="+halfSize+" r="+ray+" ";
    } else if (shape == "rect") {
         svgCode += "<rect id='"+el_name+"' width="+size+" height="+size+" ";
    }
    svgCode += "fill=black "+lut_attr+" "+stroke_opt+"/></svg>";
    aSVG.innerHTML = svgCode;
    elem_ids[el_name] = attrname;
    return aSVG;
}

class SynopticContainer {
    constructor(parent, bgnd_image = "/html/synoptic_images/bgnd_pipe_img.svg") {
        this.parent = parent;
        this.bgnd_image = bgnd_image;
        this.titles = [];
        this.images = [];
        this.img_attrs = [];
        this.alarm_attrs = [];
        this.label_attrs = [];
        this.options = [];
    }

    addElement( image = "", title = "", imgattr = "" , alarmattr = "", labelattrs = [], options = {})  {
        this.images.push(image);
        this.titles.push(title);
        this.img_attrs.push(imgattr);
        this.alarm_attrs.push(alarmattr);
        this.label_attrs.push(labelattrs);
        this.options.push(options);
    }
    
    show()  {
        aSYN = this.parent.addInfo(null,"");
        aSYN.setAttribute("class","table");
        var titleRow = '<div class="row" style="text-align:center;">';
        var imageRow = '<div class="row" style="text-align:center;">';
        var alarmRow = '<div class="row" style="text-align:center;">';
        var labelsRow = '<div class="row" style="text-align:center;">';
        for (let idx = 0; idx < this.titles.length; idx += 1) {
            let options = this.options[idx] || {};
            let o_width = 50;
            let o_height = null;
            let readonly = true;
            let lut = lut_VALVE;
            if(options.width)
                o_width = options.width;
            if(options.height)
                o_height = options.height;
            if( options.readonly == false)
                readonly = false;
            if (options.lut)
                lut = options.lut;
            //
            // Manage Title row elements
            let title = this.titles[idx];
            if (title == "") {
                title = "<br>";
            };
            titleRow += '<div class="cell" style="display: table-cell; width: '+o_width+'px;">'+title+'</div>';
            //
            // Manage Image row elements
            let image = this.images[idx];
            let boolattr = this.img_attrs[idx];
            let svg_id = "";
            let el_name = "dSynoptic" + (Object.keys(elem_ids)).length;
            if (boolattr != "") {
                svg_id += " id=" + el_name;
                svg_id += " lut=" + JSON.stringify(lut);
                elem_ids[el_name] = boolattr;
            };
            if (image == "") {
                image = "/html/synoptic_images/donkiweb_empty.png";
            };
            imageRow += '<div class="cell" style="display: table-cell; width: '+o_width+'px; height: '+o_height+'px; vertical-align: middle;">';
            if ((readonly == false) && (boolattr != "")) {
                imageRow += '<span title="'+boolattr+'" onclick="synoptic_click('+el_name+')" style="display: block; z-index: 1; cursor: pointer">';
            } else if (boolattr != "") {
                imageRow += '<span title="'+boolattr+'" style="display: block; z-index: 1;">';
            } else {
                imageRow += '<span style="display: block; z-index: 1;">';
            }
            if (image.split('.').pop() == "svg") {
                imageRow += '<object'+svg_id+' data="'+image+'" width=100% style="background-image: url('+this.bgnd_image+'); z-index: -1; pointer-events: none;"></object>';
            } else {
                imageRow += '<img src="'+image+'" width=100% height=100% style="background-image: url('+this.bgnd_image+');"/>';
            };
            imageRow += '</span></div>';
            //
            // Manage Alarms row
            alarmRow += '<div class="cell" style="display: table-cell; width: '+o_width+'px;text-align:center;">';
            let alarmattr = this.alarm_attrs[idx];
            var alrm_id = "";
            if ((typeof alarmattr === 'string') && (alarmattr !== "")) {
                el_name = "dSvgColor" + (Object.keys(elem_ids)).length;
                elem_ids[el_name] = alarmattr;
                alrm_id = " id='"+el_name+"'";
            }
            var svgCode = "<svg height='6px' width='"+o_width+"'><title>"+alarmattr+"</title>";
            var wmrg = o_width * 0.5;
            svgCode += "<rect"+alrm_id+" width='"+Math.floor(o_width-wmrg)+"px' height='4px' x='"+Math.floor(wmrg*0.5)+"px' y='1px'";
            svgCode += " fill='none' lut='"+JSON.stringify(lut_ONLYALARM)+"'/></svg>";
            alarmRow += svgCode + '</div>';
            //
            // Manage Labels row
            labelsRow += '<div class="cell" style="display: table-cell; width: '+o_width+'px;">';
            readattrs = this.label_attrs[idx];
            if (readattrs.length > 0)  {
                for (let ii = 0; ii < readattrs.length; ii += 1) {
                    let el_name = "dText" + (Object.keys(elem_ids)).length;
                    labelsRow += '<div class="cell" id='+el_name+' style="text-align:center;" title="'+readattrs[ii]+'"></div>';
                    elem_ids[el_name] = readattrs[ii];
                };
            };
            labelsRow += '</div>';
        }
        //
        titleRow += '</div>';
        imageRow += '</div>';
        alarmRow += '</div>';
        labelsRow += '</div>';
        var innerCode = titleRow + imageRow + alarmRow + labelsRow;
        aSYN.innerHTML = innerCode;
    }
}

function donkiAddSynopticContainer( parent, bgnd_image = "/html/synoptic_images/bgnd_pipe_img.svg" ) 
{
    aSYN = new SynopticContainer(parent, bgnd_image);
    return aSYN
}

function donkiAddSynopticElement( parent , image = "", title = "", imgattr = "" , alarmattr = "", labelattrs = [], options = {}) 
{
    parent.addElement(image,title,imgattr,alarmattr,labelattrs,options);
    return
}

/**
 ** positionManager class
 **/
class positionManager {
    constructor(parent, caption, attrname, options = {}) {
        this.parent = parent;
        this.caption = caption;
        this.attrname = attrname;
        this.options = options;
        this.step = 1;
        if (options.step != undefined) this.step = options.step;
        var aInfo = getTangoInfo(attrname,"attr");
        this.precision = 2;
        if (INT_TYPES.includes(aInfo.data_type)) {
            this.precision = 0;
        };
        this.sub_widgets = [];
    }
    //
    btn_callback(bname) {
        if (bname == "-") {
            var attrval = getTangoAttribute(this.parent.attrname);
            this.parent.set_enabled(false);
            writeTangoAttribute(this.parent.attrname, attrval.value - this.parent.step, false);
            setTimeout(force_update_value, 1000, this.parent.attrname);
            setTimeout(set_widget_enabled, 900, this.parent, true);
        } else if (bname == "+") {
            var attrval = getTangoAttribute(this.parent.attrname);
            this.parent.set_enabled(false);
            writeTangoAttribute(this.parent.attrname, attrval.value + this.parent.step, false);
            setTimeout(force_update_value, 1000, this.parent.attrname);
            setTimeout(set_widget_enabled, 900, this.parent, true);
        } else {
            var options = { title:"Set step value", close: true, minimize: false };
            options.width = 300;
            options.height = 60;
            options.scroll = false;
            options.resizable = false;
            options.draggable = true;
            var dialog = new LiteGUI.Dialog("SET_STEP", options);
            // Create a collection of widgets
            var widgets = new LiteGUI.Inspector();
            var prec_in = 2
            var numberWidget = widgets.addNumber("Step", this.parent.step, {precision : this.parent.precision});
            var parent = this.parent;
            dialog.add(widgets);
            //
            function applyMethod() {
                parent.step = Math.abs(numberWidget.getValue());
            }
            //
            function keyPressed( e ) {
                if (e.key === "Enter") {
                    applyMethod();
                    dialog.close();
                };
            }
            // Add some buttons
            dialog.addButton('Apply', { close: 'fade', callback: applyMethod });
            dialog.addButton('Cancel',{ close: 'fade' });
            numberWidget.addEventListener("keypress", keyPressed);
            dialog.show();
            dialog.adjustSize();
            let rect = this.getBoundingClientRect();
            dialog.setPosition(rect.left, rect.top);
        };
    };
    //
    set_enabled(value_in) {
        if (!value_in) {
            this.sub_widgets[0].disable();
        } else {
            this.sub_widgets[0].enable();
        }
        for(var i = 0; i < this.sub_widgets[1].length; ++i) {
            this.sub_widgets[1][i].disabled = !value_in;
        };
    };
    //
    show()  {
        var options = this.options;
        options.disabled = true;
        options.callback = null;
        if (options.width == undefined)
            options.width = 100;
        else
            options.width = options.width - 100;
        var aGroup = this.parent.startContainer();
        aGroup.style.display = "inline-grid";
        aGroup.style.gridTemplateColumns = "auto";
        aGroup.style.alignItems = "center";
        aGroup.style.justifyContent = "center";
        aGroup = _util_set_element_grid_style(aGroup, options);
        //
        var rownum = 1;
        if (this.caption != null) {
            let aL = donkiAddLabel(this.parent,this.caption,options);
            aL.style.gridRow = rownum;
            aL.style.gridColumn = 1 + "/span 2";
            rownum += 1;
        }
        var aRWL = donkiAddReadWriteLabel(this.parent, null, this.attrname, options);
        aRWL.style.gridRow = rownum;
        aRWL.style.gridColumn = 1;
        options.width = 100;
        options.disabled = false;
        options.callback = this.btn_callback;
        let btnLabels = ["-","step","+"];
        var aBtns = this.parent.addButtons(null, btnLabels, options);
        aBtns.style.gridRow = rownum;
        aBtns.style.gridColumn = 2;
        aBtns.parent = this;
        this.parent.endContainer();
        this.sub_widgets.push(aRWL);
        this.sub_widgets.push(aBtns.querySelectorAll("button"));
    }

}

function donkiAddPositionManager( parent, caption, attrname, options = {}) 
{
    /**
    * Widget used for read/write Tango attributes, composed by an editable
    * label (absolute writing) and two buttons (+ and -) for relative
    * writng
    * Input arguments:
    * - parent
    * - caption
    * - attrname
    * - options:
    *    - step: used for relative writing
    */
    let el_list = Object.keys(elem_ids);
    var aPM = new positionManager(parent, caption, attrname, options);
    aPM.show();
}


/**
 ** disableManager class
 **/
function donkiAddDisableManager( target_widget, attrname, disable_list = []) 
{
    /**
    * Widget used for enable/disable other widgets
    * Input arguments:
    * - target_widget
    * - attrname
    * - conditions:
    */
    let el_list = Object.keys(elem_ids);
    let el_name = "dDisableCondition" + el_list.length;
    let element = document.createElement("div");
    element.setAttribute("id", el_name);
    elem_ids[el_name] = attrname;
    element.disable_list = disable_list;
    element.target_widget = target_widget
    //
    element.apply_condition = function(value_in) {
        // TODO: check state before changing...
        element.target_widget.set_enabled(!(this.disable_list.includes(value_in)));
    };
    //
    document.body.appendChild(element);
};

